/**
 * Copyright (c) 2014-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.registry.overlay;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.exception.ConfigurationException;
import org.ow2.petals.microkernel.api.util.ConfigurationServiceUtils;
import org.ow2.petals.topology.TopologyHelper;
import org.ow2.petals.topology.generated.Container;
import org.ow2.petals.topology.generated.Topology;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;
import com.hazelcast.core.EntryEvent;
import com.hazelcast.core.EntryListener;
import com.hazelcast.core.MapEvent;

/**
 * Listener on the dynamic topology entry. Used to detect moved container from a sub-domain to another one.
 * 
 * @author Christophe DENEUX - Linagora
 */
public class DynamicTopologyListener implements EntryListener<String, Topology> {

    private final ConfigurationService configurationService;

    private final String subdomainName;

    private final LoggingUtil log;

    public DynamicTopologyListener(final ConfigurationService configurationService, final String subdomainName,
            final LoggingUtil log) {
        this.configurationService = configurationService;
        this.subdomainName = subdomainName;
        this.log = log;
    }

    /**
     * {@inheritDoc}
     */
    public void entryAdded(final EntryEvent<String, Topology> event) {
        // NOP: Only changed are managed to detect the moved containers
        this.log.call();
    }

    /**
     * {@inheritDoc}
     */
    public void entryRemoved(final EntryEvent<String, Topology> event) {
        // NOP: Only changed are managed to detect the moved containers
        this.log.call();
    }

    /**
     * {@inheritDoc}
     */
    public void entryUpdated(final EntryEvent<String, Topology> event) {
        this.log.start();
        try {
            final Topology oldTopology = event.getOldValue();
            if (oldTopology == null) {
                this.log.error("No previous topology was registered in the registry. Please contact Petals ESB Vendor !");
            } else {
                final Topology newTopology = event.getValue();
                if (this.log.isDebugEnabled()) {
                    this.log.debug("Topology updated in hazelcat cluster, old topology: " + oldTopology.toString());
                    this.log.debug("Topology updated in hazelcat cluster, new topology: " + newTopology.toString());
                }
                final List<Container> oldContainers = TopologyHelper.findSubdomain(subdomainName, oldTopology)
                        .getContainer();
                final List<Container> newContainers = TopologyHelper.findSubdomain(subdomainName, newTopology)
                        .getContainer();
                final List<Container> removedContainers = new ArrayList<Container>();
    
                final Iterator<Container> itOldContainers = oldContainers.iterator();
                while (itOldContainers.hasNext()) {
                    final Container container = itOldContainers.next();
                    if (!newContainers.contains(container)) {
                        // The current container has been removed from the sub-domain
                        removedContainers.add(container);
                    } else {
                        // The current container always exists in the sub-domain
                        newContainers.remove(container);
                    }
                }
    
                // Now, we have:
                // - all removed containers are in the list 'removedContainers',
                // - all added containers are in the list 'newContainers'.
    
                final String currentContainerName = this.configurationService.getContainerConfiguration().getName();
                final Iterator<Container> itAddedContainers = newContainers.iterator();
                while (itAddedContainers.hasNext()) {
                    try {
                        final Container container = itAddedContainers.next();
                        if (!currentContainerName.equals(container.getName())) {
                            final ContainerConfiguration containerConfig = new ContainerConfiguration();
                            ConfigurationServiceUtils.initializeContainerConfiguration(containerConfig, container,
                                    this.subdomainName);
                            this.configurationService.addContainerConfiguration(containerConfig);
                        }
                    } catch (final ConfigurationException e) {
                        this.log.warning("An error occurs adding a container detected as moved", e);
                    }
                }
    
                final Iterator<Container> itRemovedContainers = removedContainers.iterator();
                while (itRemovedContainers.hasNext()) {
                    try {
                        final Container container = itRemovedContainers.next();
                        if (!currentContainerName.equals(container.getName())) {
                            final ContainerConfiguration containerConfig = new ContainerConfiguration();
                            ConfigurationServiceUtils.initializeContainerConfiguration(containerConfig, container,
                                    this.subdomainName);
                            this.configurationService.removeContainerConfiguration(containerConfig);
                        }
                    } catch (final ConfigurationException e) {
                        this.log.warning("An error occurs adding a container detected as moved", e);
                    }
                }
            }
        } finally {
            this.log.end();
        }

    }

    /**
     * {@inheritDoc}
     */
    public void entryEvicted(final EntryEvent<String, Topology> event) {
        // NOP: Only changed are managed to detect the moved containers
        this.log.call();
    }

    /**
     * {@inheritDoc}
     */
    public void mapEvicted(MapEvent event) {
        // NOP: Only changed are managed to detect the moved containers
        this.log.call();
    }

    /**
     * {@inheritDoc}
     */
    public void mapCleared(MapEvent event) {
        // NOP: Only changed are managed to detect the moved containers
        this.log.call();
    }

}
