/**
 * Copyright (c) 2013-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.registry.overlay;

import java.io.IOException;
import java.io.StringWriter;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;

import javax.xml.namespace.QName;
import javax.xml.transform.Result;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.apache.commons.lang.NotImplementedException;
import org.apache.commons.lang.NullArgumentException;
import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.clientserverapi.jbi.messaging.registry.exception.RegistryException;
import org.ow2.petals.jbi.descriptor.original.generated.LinkType;
import org.ow2.petals.jmx.commons.PetalsEndpointRegistryServiceConstants;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.implementation.exception.InvalidPropertyException;
import org.ow2.petals.microkernel.api.implementation.exception.MissingPropertyException;
import org.ow2.petals.microkernel.api.implementation.exception.MutableImplException;
import org.ow2.petals.microkernel.api.jbi.messaging.registry.EndpointAlreadyExistsException;
import org.ow2.petals.microkernel.api.jbi.messaging.registry.EndpointDoesNotExistsException;
import org.ow2.petals.microkernel.api.jbi.messaging.registry.EndpointRegistry;
import org.ow2.petals.microkernel.api.jbi.messaging.registry.RegistryListener;
import org.ow2.petals.microkernel.api.jbi.messaging.registry.exception.DynamicTopologyNotLockedByCurrentThreadException;
import org.ow2.petals.microkernel.api.jbi.messaging.registry.exception.DynamicTopologyNotLockedException;
import org.ow2.petals.microkernel.api.jbi.servicedesc.PetalsServiceEndpoint;
import org.ow2.petals.microkernel.api.jbi.servicedesc.PetalsServiceEndpoint.EndpointType;
import org.ow2.petals.microkernel.registry.overlay.api.RegistryOverlayAPI;
import org.ow2.petals.microkernel.registry.overlay.exception.AuthenticationException;
import org.ow2.petals.microkernel.registry.overlay.exception.InvalidConfigurationException;
import org.ow2.petals.registry_overlay.api.DataModelApi;
import org.ow2.petals.registry_overlay.client.configuration.ConfigurationBuilder;
import org.ow2.petals.registry_overlay.client.configuration.ConfigurationException;
import org.ow2.petals.registry_overlay.client.configuration.generated.Configuration;
import org.ow2.petals.registry_overlay.client.configuration.generated.OverlayMember;
import org.ow2.petals.registry_overlay.client.configuration.generated.OverlayMembers;
import org.ow2.petals.topology.generated.Topology;
import org.w3c.dom.Document;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;
import com.ebmwebsourcing.easycommons.xml.DocumentBuilders;
import com.ebmwebsourcing.easycommons.xml.Transformers;
import com.hazelcast.client.HazelcastClient;
import com.hazelcast.client.config.ClientConfig;
import com.hazelcast.client.config.ClientNetworkConfig;
import com.hazelcast.config.NearCacheConfig;
import com.hazelcast.core.HazelcastInstance;
import com.hazelcast.core.ILock;
import com.hazelcast.core.IMap;
import com.hazelcast.core.MultiMap;

/**
 * The registry implementation based on the Petals Registry Overlay.
 * 
 * @author Christophe DENEUX - Linagora
 */

@Component(provides = @Interface(name = EndpointRegistry.FRACTAL_SRV_ITF_NAME, signature = EndpointRegistry.class))
public class RegistryOverlayImpl implements RegistryOverlayAPI {

    /**
     * Prefix of property names
     */
    private static final String PROPERTY_NAME_PREFIX = RegistryOverlayImpl.class.getName() + ".";
    
    /**
     * Prefix of property names configuring map cache
     */
    private static final String PROPERTY_NAME_MAP_CACHE_PREFIX = PROPERTY_NAME_PREFIX + "map-cache.";

    /**
     * Name of the property defining the max size of map cache, in endpoint number
     */
    public static final String PROPERTY_NAME_MAP_CACHE_MAX_SIZE = PROPERTY_NAME_MAP_CACHE_PREFIX + "max-size";

    /**
     * Name of the property associated to the group name
     */
    private static final String GROUP_NAME_PROPERTY_NAME = PROPERTY_NAME_PREFIX
            + RegistryOverlayAPI.RegistryItem.GROUP_NAME;
    
    /**
     * Name of the property associated to the group password
     */
    private static final String GROUP_PASSWORD_PROPERTY_NAME = PROPERTY_NAME_PREFIX
            + RegistryOverlayAPI.RegistryItem.GROUP_PASSWORD;

    /**
     * Name pattern of the property associated to an overlay member host name
     */
    private static final String MEMBER_HOST_PROPERTY_NAME_PATTERN = PROPERTY_NAME_PREFIX
            + RegistryOverlayAPI.RegistryItem.ADDRESS + ".<x>";

    /**
     * Default value of the property {@value #PROPERTY_NAME_MAP_CACHE_MAX_SIZE}
     */
    public static final int DEFAULT_MAP_CACHE_MAX_SIZE = 500;

    /**
     * The multi-map containing for each interface, the endpoints implementing them.
     */
    private MultiMap<String, String> interfacesEndpointsMultiMap;

    /**
     * The multi-map containing for each service, the endpoints implementing them.
     */
    private MultiMap<String, String> servicesEndpointsMultiMap;

    /**
     * The map containing endpoints.
     */
    private IMap<String, RegistryOverlayEndpoint> endpointsMap;

    /**
     * <p>
     * The map containing endpoint description.
     * </p>
     * <p>
     * To avoid classloader problem, the WSDL is stored as {@link String}.
     * </p>
     */
    private IMap<String, String> endpointDescriptionsMap;

    /**
     * The distributed lock managing access to the dynamic topology
     */
    private ILock dynamicTopologyLock;

    /**
     * <p>
     * The map containing dynamic topologies.
     * </p>
     * <p>
     * Note: Type is {@link IMap} to be able to register a listener on updated entries.
     * </p>
     */
    private IMap<String, Topology> topologiesMap;

    /**
     * Identifier of the listener on the dynamic topology entry
     */
    private String dynamicTopologyListener;

    /**
     * Configuration service fractal component
     */
    @Requires(name = EndpointRegistry.CONFIGURATION_FRACTAL_ITF_NAME)
    protected ConfigurationService configurationService;

    /**
     * Logging utilities
     */
    private final LoggingUtil log = new LoggingUtil(Logger.getLogger(EndpointRegistry.COMPONENT_LOGGER_NAME));

    /**
     * The registry name associated to the current domain and sub-domain
     */
    private String registryName;

    /**
     * The Hazelcast client
     */
    private HazelcastInstance hzlClient;

    private Configuration registryOverlayConfiguration = null;

    /**
     * {@inheritDoc}
     */
    public List<Map<String, Object>> getAllInternalEndpoints() {
        this.log.start();

        try {
            final List<PetalsServiceEndpoint> internalEndpoints = this.getInternalEndpoints();
            return RegistryOverlayImpl.convertEndpoints(internalEndpoints);
        } catch (final RegistryException e) {
            this.log.warning("An error ocurs getting internal endpoints", e);
            return Collections.emptyList();
        } finally {
            this.log.end();
        }
    }

    /**
     * {@inheritDoc}
     */
    public List<Map<String, Object>> getAllExternalEndpoints() {
        this.log.start();

        try {
            final List<PetalsServiceEndpoint> externalEndpoints = this.getExternalEndpoints();
            return RegistryOverlayImpl.convertEndpoints(externalEndpoints);
        } catch (final RegistryException e) {
            this.log.warning("An error ocurs getting external endpoints", e);
            return Collections.emptyList();
        } finally {
            this.log.end();
        }
    }

    /**
     * Convert a {@link PetalsServiceEndpoint} in its JMX representation. See {@link EndpointRegistry#getAllEndpoints()},
     * {@link EndpointRegistry#getAllInternalEndpoints()} or , {@link EndpointRegistry#getAllExternalEndpoints()}
     * 
     * @param serviceEndpoints
     *            A list of service endpoints to convert
     * @return The service endpoints converted. Not <code>null</code>. Can be empty.
     */
    private static List<Map<String, Object>> convertEndpoints(final List<PetalsServiceEndpoint> serviceEndpoints) {

        final List<Map<String, Object>> convertedEndpoints = new ArrayList<Map<String, Object>>();
        for (final PetalsServiceEndpoint serviceEndpoint : serviceEndpoints) {
            final Map<String, Object> endpointMap = new HashMap<String, Object>();
            endpointMap.put(PetalsEndpointRegistryServiceConstants.Endpoint.COMPONENT_NAME, serviceEndpoint
                    .getLocation().getComponentName());
            endpointMap.put(PetalsEndpointRegistryServiceConstants.Endpoint.CONTAINER_NAME, serviceEndpoint
                    .getLocation().getContainerName());
            endpointMap.put(PetalsEndpointRegistryServiceConstants.Endpoint.ENDPOINT_NAME,
                    serviceEndpoint.getEndpointName());
            endpointMap.put(PetalsEndpointRegistryServiceConstants.Endpoint.SERVICE_NAME, serviceEndpoint
                    .getServiceName().toString());

            final EndpointType endpointType = serviceEndpoint.getType();
            if (endpointType.equals(EndpointType.INTERNAL)) {
                endpointMap.put(PetalsEndpointRegistryServiceConstants.Endpoint.TYPE,
                        PetalsEndpointRegistryServiceConstants.EndpointType.INTERNAL);
            } else if (endpointType.equals(EndpointType.EXTERNAL)) {
                endpointMap.put(PetalsEndpointRegistryServiceConstants.Endpoint.TYPE,
                        PetalsEndpointRegistryServiceConstants.EndpointType.EXTERNAL);
            }

            final QName[] interfaces = serviceEndpoint.getInterfaces();
            final List<String> convertedInterfaceNames = new ArrayList<String>();
            if (interfaces.length > 0) {
                for (final QName interfaceName : interfaces) {
                    convertedInterfaceNames.add(interfaceName.toString());
                }
            } else {
                convertedInterfaceNames.add(PetalsEndpointRegistryServiceConstants.Interface.UNDEFINED);
            }
            endpointMap.put(PetalsEndpointRegistryServiceConstants.Endpoint.INTERFACE_NAMES,
                    convertedInterfaceNames.toArray(new String[convertedInterfaceNames.size()]));

            convertedEndpoints.add(endpointMap);
        }

        return convertedEndpoints;
    }

    /**
     * {@inheritDoc}
     */
    public List<Map<String, Object>> getAllEndpoints() {

        this.log.start();

        try {
            final List<Map<String, Object>> allEndpointsMap = this.getAllExternalEndpoints();
            allEndpointsMap.addAll(this.getAllInternalEndpoints());

            return allEndpointsMap;
        } finally {
            this.log.end();
        }
    }

    /**
     * {@inheritDoc}
     */
    public String getDescription(final String serviceName, final String endpointName) throws RegistryException {

        if (this.log.isDebugEnabled()) {
            this.log.start("with params serviceName = " + serviceName + ", endpointName = " + endpointName);
        }

        try {
            final PetalsServiceEndpoint svcEp = this.getEndpoint(QName.valueOf(serviceName), endpointName);
            if (svcEp != null) {
                final Document description = svcEp.getDescription();
                final StringWriter sw = new StringWriter();
                final StreamResult sr = new StreamResult(sw);
                final Transformer transformer = Transformers.takeTransformer();
                try {
                    transformer.transform(new DOMSource(description), sr);
                    return sw.toString();

                } catch (final TransformerException e) {
                    throw new RegistryException(e);
                } finally {
                    Transformers.releaseTransformer(transformer);
                    try {
                        sw.close();
                    } catch (final IOException e) {
                        // NOP: Should not occur
                        this.log.warning("An error occurs closing a StringWriter. Skipped.", e);
                    }
                }
            } else {
                return null;
            }
        } catch (final RuntimeException e) {
            throw new RegistryException(e);
        } catch (final RegistryException e) {
            throw new RegistryException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    public List<RegistryListener> getListeners() {
        // TODO Auto-generated method stub
        return null;
    }

    /**
     * {@inheritDoc}
     */
    public List<PetalsServiceEndpoint> getInternalEndpoints() throws RegistryException {
        try {
            return this.getEndpoints(EndpointType.INTERNAL);
        } catch (final RuntimeException e) {
            throw new RegistryException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    public List<PetalsServiceEndpoint> getExternalEndpoints() throws RegistryException {
        try {
            return this.getEndpoints(EndpointType.EXTERNAL);
        } catch (final RuntimeException e) {
            throw new RegistryException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    public List<PetalsServiceEndpoint> getEndpoints() throws RegistryException {
        try {
            return new ArrayList<PetalsServiceEndpoint>(this.endpointsMap.values());
        } catch (final RuntimeException e) {
            throw new RegistryException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    public PetalsServiceEndpoint activateEndpoint(final QName serviceName, final String endpointName,
            final List<QName> interfaces, final Document description, final PetalsServiceEndpoint address,
            final boolean canOverwrite)
            throws EndpointAlreadyExistsException, RegistryException {

        final RegistryOverlayEndpoint endpoint = new RegistryOverlayEndpoint(endpointName, serviceName, interfaces,
                address.getLocation(), this);

        this.registerEndpoint(endpoint, description, canOverwrite);

        return endpoint;
    }

    /**
     * 
     * @param serviceEndpoint
     * @param description
     * @param canOverwrite
     *            If set to <code>true</code>, if the endpoint already exists, it will be overwritten, otherwise the
     *            error {@link EndpointAlreadyExistsException} is thrown
     * @throws EndpointAlreadyExistsException
     *             The endpoint already exists, and <code>canOverwrite</code> is not set to <code>true</code>.
     * @throws RegistryException
     */
    private void registerEndpoint(final RegistryOverlayEndpoint serviceEndpoint, final Document description,
            final boolean canOverwrite)
            throws RegistryException {

        if (serviceEndpoint == null) {
            throw new RegistryException("Can not register a null service endpoint");
        }

        final String endpointName = serviceEndpoint.getEndpointName();
        final QName serviceName = serviceEndpoint.getServiceName();

        // If the service endpoint cannot be overwritten, We check if it already exists.
        if (!canOverwrite && this.getEndpoint(serviceName, endpointName) != null) {
            throw new EndpointAlreadyExistsException(endpointName, serviceName);
        }

        try {
            // For each interface
            for (final QName itf : serviceEndpoint.getInterfaces()) {
                // we register the endpoint as implementing the interface
                this.interfacesEndpointsMultiMap.put(itf.toString(), endpointName);
            }

            // We register the endpoint as implementing the service name
            this.servicesEndpointsMultiMap.put(serviceName.toString(), endpointName);

            // We register the endpoint
            this.endpointsMap.set(endpointName, serviceEndpoint);

            // For performance reasons, the description is not saved as part of the service endpoint. It is retrieved
            // only if needed through ServiceEndpoint.getDescription(). See RegistryOverlayEndpoint.
            if (description != null) {
                final StringWriter writer = new StringWriter();
                final Result streamResult = new StreamResult(writer);
                final Transformer transformer = Transformers.takeTransformer();
                try {
                    transformer.transform(new DOMSource(description), streamResult);
                    this.endpointDescriptionsMap.set(endpointName, writer.toString());
                } catch (final TransformerException e) {
                    throw new RegistryException(e);
                } finally {
                    Transformers.releaseTransformer(transformer);
                    try {
                        writer.close();
                    } catch (final IOException e) {
                        // NOP: Should not occur
                        this.log.warning("An error occurs closing a StringWriter. Skipped.", e);
                    }
                }
            }
        } catch (final RuntimeException e) {
            throw new RegistryException(e);
        }

    }

    /**
     * {@inheritDoc}
     */
    public void deactivateEndpoint(final String endpointName, final QName serviceName) throws RegistryException {

        if ((endpointName == null) || (serviceName == null)) {
            throw new RegistryException("Can not deactivate null endpoint");
        }

        try {
            final PetalsServiceEndpoint svcEp = this.endpointsMap.remove(endpointName);
            this.endpointDescriptionsMap.remove(endpointName);

            // We check if the service-endpoint already exists.
            if (svcEp == null) {
                throw new EndpointDoesNotExistsException(endpointName, serviceName);
            } else {

                for (final QName itf : svcEp.getInterfaces()) {
                    this.interfacesEndpointsMultiMap.get(itf.toString()).remove(endpointName);
                }

                this.servicesEndpointsMultiMap.get(serviceName.toString()).remove(endpointName);
            }
        } catch (final RuntimeException e) {
            throw new RegistryException(e);
        }

    }

    /**
     * {@inheritDoc}
     */
    public void registerExternalEndpoint(final javax.jbi.servicedesc.ServiceEndpoint externalSvcEp)
            throws RegistryException {

        if (this.log.isDebugEnabled()) {
            this.log.start("with params externalSvcEp = " + externalSvcEp);
        }

        try {
            if (externalSvcEp == null) {
                throw new RegistryException("Can not register a null external endpoint");
            }
            if (externalSvcEp instanceof PetalsServiceEndpoint) {
                this.registerEndpoint(new RegistryOverlayEndpoint((PetalsServiceEndpoint) externalSvcEp, this),
                        ((PetalsServiceEndpoint) externalSvcEp).getDescription(), false);
            } else {
                this.log.warning("Trying to register a bad external service endpoint");
                throw new RegistryException("Bad external service endpoint type");
            }
        } catch (final RuntimeException e) {
            throw new RegistryException(e);
        } finally {
            this.log.end();
        }

    }

    /**
     * {@inheritDoc}
     */
    public void deregisterExternalEndpoint(final String endpointName, final QName serviceName) throws RegistryException {
        this.deactivateEndpoint(endpointName, serviceName);
    }

    /**
     * {@inheritDoc}
     */
    public void registerConnection(final QName consInterface, final QName provService, final String provEndpoint)
            throws RegistryException {
        throw new NotImplementedException();

    }

    /**
     * {@inheritDoc}
     */
    public void registerConnection(final QName consService, final String consEndpoint, final QName provService,
            final String provEndpoint)
            throws RegistryException {
        throw new NotImplementedException();

    }

    /**
     * {@inheritDoc}
     */
    public void deregisterConnection(final QName consInterface, final QName provService, final String provEndpoint)
            throws RegistryException {

        throw new NotImplementedException();
    }

    /**
     * {@inheritDoc}
     */
    public void deregisterConnection(final QName consService, final String consEndpoint, final QName provService,
            final String provEndpoint) throws RegistryException {

        throw new NotImplementedException();
    }

    /**
     * {@inheritDoc}
     */
    public Document getEndpointDescriptorForEndpoint(final PetalsServiceEndpoint endpoint) throws RegistryException {

        if (this.log.isDebugEnabled()) {
            this.log.start("with params endpoint = " + endpoint);
        }

        if (endpoint == null) {
            throw new RegistryException("Null endpoint");
        }

        try {
            Document result = endpoint.getDescription();
            if (result == null) {

                // try to get the endpoint from the registry
                final PetalsServiceEndpoint retrievedSvcEp = this.getEndpoint(endpoint.getServiceName(),
                        endpoint.getEndpointName());
                if (retrievedSvcEp != null) {
                    result = retrievedSvcEp.getDescription();
                }
            }
            return result;
        } catch (final RuntimeException e) {
            throw new RegistryException(e);
        } finally {
            this.log.end();
        }
    }

    /**
     * {@inheritDoc}
     */
    public QName[] getInterfacesForEndpoint(final PetalsServiceEndpoint svcEp) {

        if (this.log.isDebugEnabled()) {
            this.log.start("with params endpoint = " + svcEp);
        }

        if (svcEp == null) {
            throw new NullArgumentException("endpoint");
        }

        try {
            final QName[] existingItfs = svcEp.getInterfaces();
            if (existingItfs == null) {
                final PetalsServiceEndpoint retrievedSvcEp = this.endpointsMap.get(svcEp.getEndpointName());
                if (retrievedSvcEp != null) {
                    return retrievedSvcEp.getInterfaces();
                } else {
                    return null;
                }
            } else {
                return existingItfs;
            }
        } catch (final RuntimeException e) {
            this.log.warning("An error occurs getting interfaces of service", e);
            return null;
        } finally {
            this.log.end();
        }
    }

    /**
     * {@inheritDoc}
     */
    public PetalsServiceEndpoint[] getInternalEndpointsForInterface(final QName interfaceName, final LinkType linktype)
            throws RegistryException {

        if (this.log.isDebugEnabled()) {
            this.log.start("with params interfaceName = " + interfaceName + ", linktype = " + linktype);
        }

        try {
            return this.getEndpointsForInterface(interfaceName, EndpointType.INTERNAL);
        } catch (final RuntimeException e) {
            throw new RegistryException(e);
        } finally {
            this.log.end();
        }

    }

    /**
     * Queries the registry for endpoints that implements the given interface name
     * 
     * @param interfaceName
     *            qualified name of interface implemented by the endpoints; can be null
     * @return an array of available endpoints for the specified interface name; returns all available endpoints if
     *         interface name is null; must be non null, may be empty
     */
    private PetalsServiceEndpoint[] getEndpointsForInterface(final QName interfaceName, final EndpointType type)
            throws RegistryException {

        final List<PetalsServiceEndpoint> result = new ArrayList<PetalsServiceEndpoint>();

        try {
            if (interfaceName != null) {
                final Collection<String> endpoints = this.interfacesEndpointsMultiMap.get(interfaceName.toString());
                final Iterator<String> itEndpoints = endpoints.iterator();
                while (itEndpoints.hasNext()) {
                    final String endpointName = itEndpoints.next();
                    final RegistryOverlayEndpoint serviceEndpoint = this.endpointsMap.get(endpointName);
                    if (serviceEndpoint != null && serviceEndpoint.getType() != null
                            && serviceEndpoint.getType().equals(type)) {

                        serviceEndpoint.setRegistryOverlay(this);
                        result.add(serviceEndpoint);
                    }
                }
            } else {
                final Iterator<RegistryOverlayEndpoint> itSvcEndpoints = this.endpointsMap.values().iterator();
                while (itSvcEndpoints.hasNext()) {
                    final RegistryOverlayEndpoint serviceEndpoint = itSvcEndpoints.next();
                    if (serviceEndpoint != null && serviceEndpoint.getType() != null
                            && serviceEndpoint.getType().equals(type)) {

                        serviceEndpoint.setRegistryOverlay(this);
                        result.add(serviceEndpoint);
                    }
                }
            }
            return result.toArray(new PetalsServiceEndpoint[result.size()]);
        } catch (final RuntimeException e) {
            throw new RegistryException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    public PetalsServiceEndpoint[] getInternalEndpointsForService(final QName serviceName, final LinkType linktype)
            throws RegistryException {

        if (this.log.isDebugEnabled()) {
            this.log.start("with params serviceName = " + serviceName + ", linktype = " + linktype);
        }

        try {
            return this.getEndpointsForService(serviceName, EndpointType.INTERNAL);
        } catch (final RuntimeException e) {
            throw new RegistryException(e);
        } finally {
            this.log.end();
        }

    }

    /**
     * {@inheritDoc}
     */
    public PetalsServiceEndpoint[] getExternalEndpointsForInterface(final QName interfaceName) throws RegistryException {

        if (this.log.isDebugEnabled()) {
            this.log.start("with params interfaceName = " + interfaceName);
        }

        try {
            return this.getEndpointsForInterface(interfaceName, EndpointType.EXTERNAL);
        } catch (final RuntimeException e) {
            throw new RegistryException(e);
        } finally {
            this.log.end();
        }
    }

    /**
     * {@inheritDoc}
     */
    public PetalsServiceEndpoint[] getExternalEndpointsForService(final QName serviceName) throws RegistryException {

        if (this.log.isDebugEnabled()) {
            this.log.start("with params serviceName = " + serviceName);
        }

        try {
            return this.getEndpointsForService(serviceName, EndpointType.EXTERNAL);
        } catch (final RuntimeException e) {
            throw new RegistryException(e);
        } finally {
            this.log.end();
        }
    }

    /**
     * {@inheritDoc}
     */
    public PetalsServiceEndpoint getEndpoint(final QName service, final String endpointName) throws RegistryException {
        if (this.log.isDebugEnabled()) {
            this.log.start("with params serviceName = " + service.toString() + ", endpointName = " + endpointName);
        }

        try {
            final RegistryOverlayEndpoint serviceEndpoint = this.endpointsMap.get(endpointName);
            if (serviceEndpoint != null) {
                // We check that the endpoint implements the expected service
                if (serviceEndpoint.getServiceName().equals(service)) {

                    serviceEndpoint.setRegistryOverlay(this);
                    return serviceEndpoint;
                } else {
                    return null;
                }
            } else {
                return null;
            }
        } catch (final RuntimeException e) {
            throw new RegistryException(e);
        } finally {
            this.log.end();
        }
    }

    /**
     * {@inheritDoc}
     */
    public void removeAllLocalEndpoints() throws RegistryException {
        this.log.start();

        final String localContainerName = this.configurationService.getContainerConfiguration().getName();

        try {
            final Iterator<RegistryOverlayEndpoint> endpoints = this.endpointsMap.values().iterator();
            while (endpoints.hasNext()) {
                final PetalsServiceEndpoint endpoint = endpoints.next();
                if (endpoint.getLocation().getContainerName().equals(localContainerName)) {
                    this.deactivateEndpoint(endpoint.getEndpointName(), endpoint.getServiceName());
                }
            }
        } catch (final RuntimeException e) {
            throw new RegistryException(e);
        }
    }

    @Lifecycle(step = Step.START)
    public void start() throws RegistryException {

        this.log.start();

        try {
            final String domainName = this.configurationService.getDomainConfiguration().getName();
            if (domainName == null) {
                throw new InvalidConfigurationException("The domain name is missing");
            }
            final String subdomainName = this.configurationService.getSubDomainConfiguration().getName();
            if (subdomainName == null) {
                throw new InvalidConfigurationException("The sub-domain name is missing");
            }
            this.registryName = DataModelApi.getRegistryName(domainName, subdomainName);

            final List<Object> extraParameters = this.configurationService.getSubDomainConfiguration()
                    .getExtraParameters();
            if (extraParameters != null) {
                try {
                    for (final Object extraParameter : extraParameters) {
                        final Configuration cfg = ConfigurationBuilder
                                .createConfiguration((org.w3c.dom.Node) extraParameter);
                        if (cfg != null) {
                            this.registryOverlayConfiguration = cfg;
                            break;
                        }
                    }
                } catch (final ConfigurationException e) {
                    throw new InvalidConfigurationException(e);
                }
                if (this.registryOverlayConfiguration == null) {
                    throw new InvalidConfigurationException("Missing configuration");
                }
            } else {
                throw new InvalidConfigurationException("Missing configuration");
            }

            if (this.registryOverlayConfiguration.getGroupName() == null
                    || this.registryOverlayConfiguration.getGroupName().isEmpty()) {
                throw new InvalidConfigurationException("The group name is missing");
            }
            if (this.registryOverlayConfiguration.getGroupPassword() == null
                    || this.registryOverlayConfiguration.getGroupPassword().isEmpty()) {
                throw new InvalidConfigurationException("The group password is missing");
            }
            if (this.registryOverlayConfiguration.getOverlayMembers() == null
                    || this.registryOverlayConfiguration.getOverlayMembers().getOverlayMember() == null
                    || this.registryOverlayConfiguration.getOverlayMembers().getOverlayMember().isEmpty()) {
                throw new InvalidConfigurationException("No overlay member declared");
            }

            if (this.log.isConfigEnabled()) {
                this.log.config("Registry name: " + this.registryName);
                this.log.config("Group name: " + this.registryOverlayConfiguration.getGroupName());
                this.log.config("Group pwd: ****");
            }

            // Connect to the Hazelcast cluster
            final ClientConfig cltCfg = new ClientConfig();

            cltCfg.getGroupConfig().setName(this.registryOverlayConfiguration.getGroupName());
            cltCfg.getGroupConfig().setPassword(this.registryOverlayConfiguration.getGroupPassword());
            final ClientNetworkConfig networkConfig = new ClientNetworkConfig();
            cltCfg.setNetworkConfig(networkConfig);
            for (final OverlayMember overlayMember : this.registryOverlayConfiguration.getOverlayMembers()
                    .getOverlayMember()) {
                final String host = overlayMember.getValue();
                final BigInteger port = overlayMember.getPort();
                if (host == null || host.isEmpty()) {
                    throw new InvalidConfigurationException("The hostname of an overlay member is missing");
                }
                final String address = host + ":" + (port == null ? DEFAULT_MEMBER_PORT : port.toString());
                if (this.log.isConfigEnabled()) {
                    this.log.config("Member: " + address);
                }
                networkConfig.addAddress(address);
            }

            // Enable near cache
            this.enableNearCache(cltCfg);

            // We use the Petals ESB container classloader
            cltCfg.setClassLoader(Thread.currentThread().getContextClassLoader());

            try {
                this.hzlClient = HazelcastClient.newHazelcastClient(cltCfg);

                // When no more clients associated to a dedicated topology is connected, all data relative to the
                // topology must be removed from the HZL cluster. To manage nominal client shutdown and unexpected
                // client shutdown, the HZL cluster will detect itself when clients are disconnected. When no more
                // client is connected, the HZL cluster will clean up topology information.
                // That's why we register the client, as <ip-address:port> in a distributed map. The HZL cluster will
                // deregister the client on disconnection event.
                final IMap<String, String> clients = DataModelApi.getClientsMap(this.hzlClient);
                clients.set(this.hzlClient.getLocalEndpoint().getUuid(), this.registryName);

                // Distributed objects about endpoints
                this.interfacesEndpointsMultiMap = DataModelApi.getInterfacesEndpointsMultiMap(this.hzlClient,
                        this.registryName);
                this.servicesEndpointsMultiMap = DataModelApi.getServicesEndpointsMultiMap(this.hzlClient,
                        this.registryName);
                this.endpointsMap = DataModelApi.getEndpointsMap(this.hzlClient, this.registryName);
                this.endpointDescriptionsMap = DataModelApi.getEndpointDescriptionsMap(this.hzlClient,
                        this.registryName);

                // Distributed objects about dynamic topology
                this.dynamicTopologyLock = DataModelApi.getTopologiesMapLock(this.hzlClient);
                this.topologiesMap = DataModelApi.getTopologiesMap(this.hzlClient);
                
                this.dynamicTopologyListener = this.topologiesMap.addEntryListener(new DynamicTopologyListener(
                        this.configurationService, subdomainName, this.log), this.registryName, true);

            } catch (final RuntimeException e) {
                if (e.getCause() != null && e.getCause() instanceof com.hazelcast.client.AuthenticationException) {
                    throw new AuthenticationException();
                } else {
                    throw new RegistryException(e);
                }
            }
        } catch (final RegistryException e) {
            this.log.error("An error occurs starting the registry client", e);
            throw e;
        } catch (final Throwable e) {
            this.log.error("An error occurs starting the registry client", e);
            throw new RegistryException(e);
        } finally {
            this.log.end();
        }
    }

    /**
     * Enable NearCache for all distributed map
     * 
     * @param cltCfg
     *            The Hazelcast client configuration
     */
    private final void enableNearCache(final ClientConfig cltCfg) {

        final ContainerConfiguration containerConfiguration = this.configurationService.getContainerConfiguration();
        final Map<String, String> extraConfProperties = containerConfiguration.getExtraConfiguration();
        int maxSize;
        if (extraConfProperties != null) {
            final String maxSizeStr = extraConfProperties.get(PROPERTY_NAME_MAP_CACHE_MAX_SIZE);
            if (maxSizeStr != null && !maxSizeStr.trim().isEmpty()) {
                try {
                    maxSize = Integer.parseInt(maxSizeStr);
                    this.log.config("Max size of map caches: " + maxSize);
                } catch (final NumberFormatException e) {
                    this.log.warning(String.format(
                            "Invalid value for the max size of map cache: '%s'. Default value used: '%s'", maxSizeStr,
                            DEFAULT_MAP_CACHE_MAX_SIZE));
                    maxSize = DEFAULT_MAP_CACHE_MAX_SIZE;
                }
            } else {
                this.log.config("The max size of map cache is not defined. Default value used: "
                        + DEFAULT_MAP_CACHE_MAX_SIZE);
                maxSize = DEFAULT_MAP_CACHE_MAX_SIZE;
            }
        } else {
            this.log.config("No extra configuration defined for the registry overlay client implementation. Default values used");
            maxSize = DEFAULT_MAP_CACHE_MAX_SIZE;
        }

        final NearCacheConfig cltNearCacheEndpointsCfg = new NearCacheConfig();
        cltNearCacheEndpointsCfg.setName(DataModelApi.getEndpointsMapName(this.registryName));
        cltNearCacheEndpointsCfg.setMaxSize(maxSize);
        cltNearCacheEndpointsCfg.setInvalidateOnChange(Boolean.TRUE.booleanValue());
        cltNearCacheEndpointsCfg.setEvictionPolicy("LRU");
        cltCfg.addNearCacheConfig(cltNearCacheEndpointsCfg);

        final NearCacheConfig cltNearCacheEndpointDescriptionsCfg = new NearCacheConfig();
        cltNearCacheEndpointDescriptionsCfg.setName(DataModelApi.getEndpointDescriptionsMapName(this.registryName));
        cltNearCacheEndpointDescriptionsCfg.setMaxSize(maxSize);
        cltNearCacheEndpointDescriptionsCfg.setInvalidateOnChange(Boolean.TRUE.booleanValue());
        cltNearCacheEndpointDescriptionsCfg.setEvictionPolicy("LRU");
        cltCfg.addNearCacheConfig(cltNearCacheEndpointDescriptionsCfg);

        final NearCacheConfig cltNearCacheInterfaceEndpointsCfg = new NearCacheConfig();
        cltNearCacheInterfaceEndpointsCfg.setName(DataModelApi.getInterfacesEndpointsMultiMapName(this.registryName));
        cltNearCacheInterfaceEndpointsCfg.setMaxSize(maxSize);
        cltNearCacheInterfaceEndpointsCfg.setInvalidateOnChange(Boolean.TRUE.booleanValue());
        cltNearCacheInterfaceEndpointsCfg.setEvictionPolicy("LRU");
        cltCfg.addNearCacheConfig(cltNearCacheInterfaceEndpointsCfg);

        final NearCacheConfig cltNearCacheServiceEndpointsCfg = new NearCacheConfig();
        cltNearCacheServiceEndpointsCfg.setName(DataModelApi.getServicesEndpointsMultiMapName(this.registryName));
        cltNearCacheServiceEndpointsCfg.setMaxSize(maxSize);
        cltNearCacheServiceEndpointsCfg.setInvalidateOnChange(Boolean.TRUE.booleanValue());
        cltNearCacheServiceEndpointsCfg.setEvictionPolicy("LRU");
        cltCfg.addNearCacheConfig(cltNearCacheServiceEndpointsCfg);

        final NearCacheConfig cltNearCacheTopologieCfg = new NearCacheConfig();
        cltNearCacheTopologieCfg.setName(DataModelApi.getTopologiesMapName());
        cltNearCacheTopologieCfg.setMaxSize(maxSize);
        cltNearCacheTopologieCfg.setInvalidateOnChange(Boolean.TRUE.booleanValue());
        cltNearCacheTopologieCfg.setEvictionPolicy("LRU");
        cltCfg.addNearCacheConfig(cltNearCacheTopologieCfg);
    }

    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.start();

        try {
            if (this.hzlClient != null) {
                this.topologiesMap.removeEntryListener(this.dynamicTopologyListener);
                this.hzlClient.shutdown();
            }
        } catch (final Exception e) {
            this.log.warning("An error occured while stopping the registry client", e);
        }

        this.log.end();
    }

    /**
     * {@inheritDoc}
     */
    public Properties getConfiguration(final String providedTopologyPassphrase) {
        final Properties result = new Properties();
        if (this.registryOverlayConfiguration != null) {
            final boolean isSecurityOk = this.configurationService
                    .isSecurityTopologyPassphraseOk(providedTopologyPassphrase);
            if (this.log.isDebugEnabled()) {
                if (isSecurityOk) {
                    this.log.debug("Security validated. Sensible information will be returned.");
                } else {
                    this.log.debug("Security not validated. Sensible information will NOT be returned.");
                }
            }

            result.setProperty(PROPERTY_NAME_PREFIX + RegistryOverlayAPI.RegistryItem.GROUP_NAME,
                    this.registryOverlayConfiguration.getGroupName());
            if (isSecurityOk) {
                result.setProperty(PROPERTY_NAME_PREFIX + RegistryOverlayAPI.RegistryItem.GROUP_PASSWORD,
                        this.registryOverlayConfiguration.getGroupPassword());
            }

            int addressNb = 0;
            for (final OverlayMember member : this.registryOverlayConfiguration.getOverlayMembers().getOverlayMember()) {
                final String host = member.getValue();
                final BigInteger port = member.getPort();
                if (host != null && !host.isEmpty()) {
                    if (port == null) {
                        result.setProperty(PROPERTY_NAME_PREFIX + RegistryOverlayAPI.RegistryItem.ADDRESS + "."
                                + addressNb++, host + ":" + DEFAULT_MEMBER_PORT);
                    } else {
                        result.setProperty(PROPERTY_NAME_PREFIX + RegistryOverlayAPI.RegistryItem.ADDRESS + "."
                                + addressNb++, host + ":" + port.toString());
                    }
                }
            }
        }
        return result;
    }

    /**
     * {@inheritDoc}
     */
    public Document getTopologyPartConfiguration(final Properties properties) throws MutableImplException {
        
        if (properties != null) {
            
            final Configuration xmlConfiguration = new Configuration();
    
            // Group name
            final String groupName = properties.getProperty(GROUP_NAME_PROPERTY_NAME);
            if (groupName != null) {
                xmlConfiguration.setGroupName(groupName);
            } else {
                throw new MissingPropertyException(GROUP_NAME_PROPERTY_NAME);
            }

            final String groupPwd = properties.getProperty(GROUP_PASSWORD_PROPERTY_NAME);
            if (groupPwd != null) {
                xmlConfiguration.setGroupPassword(groupPwd);
            } else {
                throw new MissingPropertyException(GROUP_PASSWORD_PROPERTY_NAME);
            }

            xmlConfiguration.setOverlayMembers(new OverlayMembers());
            boolean isAddressFound = false;
            for (final String propertyName : properties.stringPropertyNames()) {
                if (propertyName.startsWith(PROPERTY_NAME_PREFIX + RegistryOverlayAPI.RegistryItem.ADDRESS)) {
                    isAddressFound = true;
                    final String address = properties.getProperty(propertyName);
                    if (address != null && !address.isEmpty()) {
                        final String[] addressTokens = address.split(":");
                        if (addressTokens.length > 2) {
                            throw new InvalidPropertyException(propertyName, address);
                        } else {
                            final String host = addressTokens[0];
                            if (host.isEmpty()) {
                                throw new InvalidPropertyException(propertyName, address);
                            } else {
                                final OverlayMember member = new OverlayMember();
                                xmlConfiguration.getOverlayMembers().getOverlayMember().add(member);
                                member.setValue(host);

                                if (addressTokens.length == 2) {
                                    final String portStr = addressTokens[1];
                                    if (portStr != null && !portStr.isEmpty()) {
                                        member.setPort(new BigInteger(portStr));
                                    }
                                }
                            }
                        }
                    } else {
                        throw new InvalidPropertyException(propertyName, address);
                    }
                }
            }
            if (!isAddressFound) {
                throw new MissingPropertyException(MEMBER_HOST_PROPERTY_NAME_PATTERN);
            }

            try {
                final Document document = DocumentBuilders.newDocument();
                ConfigurationBuilder.marshallConfiguration(xmlConfiguration, document);
                return document;
            } catch (final ConfigurationException e) {
                throw new MutableImplException(e);
            }
        } else {
            return null;
        }
    }

    /**
     * Get endpoints of type '<code>type</code>'
     * 
     * @param type
     *            The type that the endpoints must implement
     * @return The endpoints matching '<code>type</code>'
     */
    private List<PetalsServiceEndpoint> getEndpoints(final EndpointType type) {

        final List<PetalsServiceEndpoint> result = new ArrayList<PetalsServiceEndpoint>(this.endpointsMap.size());

        final Iterator<RegistryOverlayEndpoint> itEndpoints = this.endpointsMap.values().iterator();
        while (itEndpoints.hasNext()) {
            final RegistryOverlayEndpoint serviceEndpoint = itEndpoints.next();
            if (serviceEndpoint.getType() != null && serviceEndpoint.getType().equals(type)) {

                serviceEndpoint.setRegistryOverlay(this);
                result.add(serviceEndpoint);
            }
        }
        return result;
    }

    /**
     * Get endpoints of type '<code>type</code>' implementing the service '<code>serviceName</code>'
     * 
     * @param serviceName
     *            The service that the endpoints must implement
     * @param type
     *            The type that the endpoints must implement
     * @return The endpoints matching '<code>serviceName</code>' and '<code>type</code>'
     * @throws RegistryException
     *             A parameter is invalid
     */
    private PetalsServiceEndpoint[] getEndpointsForService(final QName serviceName, final EndpointType type)
            throws RegistryException {

        if (serviceName == null) {
            throw new RegistryException("No serviceName found");
        }

        try {
            final Collection<String> endpoints = this.servicesEndpointsMultiMap.get(serviceName.toString());
            final List<RegistryOverlayEndpoint> result = new ArrayList<RegistryOverlayEndpoint>(endpoints.size());

            final Iterator<String> itEndpoints = endpoints.iterator();
            while (itEndpoints.hasNext()) {
                final String endpointName = itEndpoints.next();
                final RegistryOverlayEndpoint serviceEndpoint = this.endpointsMap.get(endpointName);
                if (serviceEndpoint != null && serviceEndpoint.getType() != null
                        && serviceEndpoint.getType().equals(type)) {

                    serviceEndpoint.setRegistryOverlay(this);
                    result.add(serviceEndpoint);
                }
            }
            return result.toArray(new PetalsServiceEndpoint[result.size()]);
        } catch (final RuntimeException e) {
            throw new RegistryException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    public String retrieveDescription(final String endpointName) {
        return this.endpointDescriptionsMap.get(endpointName);
    }

    /**
     * {@inheritDoc}
     */
    public boolean lockDynamicTopology() throws InterruptedException {
        final ContainerConfiguration containerCfg = this.configurationService.getContainerConfiguration();
        return this.dynamicTopologyLock.tryLock(containerCfg.getTopologyDynamicLockWaitTime(), TimeUnit.SECONDS);
    }

    /**
     * {@inheritDoc}
     */
    public void unlockDynamicTopology() throws DynamicTopologyNotLockedException,
            DynamicTopologyNotLockedByCurrentThreadException {
        if (this.dynamicTopologyLock.isLocked()) {
            if (this.dynamicTopologyLock.isLockedByCurrentThread()) {
                this.dynamicTopologyLock.unlock();
            } else {
                throw new DynamicTopologyNotLockedByCurrentThreadException();
            }
        } else {
            throw new DynamicTopologyNotLockedException();
        }
    }

    /**
     * {@inheritDoc}
     */
    public void setDynamicTopology(final Topology topology) throws DynamicTopologyNotLockedException,
            DynamicTopologyNotLockedByCurrentThreadException {
        this.log.start();
        try {
            if (this.dynamicTopologyLock.isLocked()) {
                if (this.dynamicTopologyLock.isLockedByCurrentThread()) {
                    // It is required to use IMap.put() instead of IMap.set() because IMap.set() breaks the contract of
                    // EntryListener firing an EntryEvent with a null oldValue. A non-null oldValue is required by
                    // DynamicTopologyListener.entryUpdated() to update the topology changes.
                    this.topologiesMap.put(this.registryName, topology);
                } else {
                    throw new DynamicTopologyNotLockedByCurrentThreadException();
                }
            } else {
                throw new DynamicTopologyNotLockedException();
            }
        } finally {
            this.log.end();
        }
    }

    /**
     * {@inheritDoc}
     */
    public Topology getDynamicTopology() {
        try {
            return this.topologiesMap.get(this.registryName);
        } finally {
            this.log.end();
        }
    }
}
