/**
 * Copyright (c) 2013-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.registry.overlay;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;

import java.util.Set;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.util.ConfigurationServiceUtils;
import org.ow2.petals.microkernel.configuration.ConfigurationTestUtils;
import org.ow2.petals.topology.generated.Container;
import org.ow2.petals.topology.generated.JmxService;
import org.ow2.petals.topology.generated.Subdomain;
import org.ow2.petals.topology.generated.Topology;
import org.ow2.petals.topology.generated.TransportService;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;
import com.hazelcast.client.HazelcastClient;
import com.hazelcast.client.config.ClientConfig;
import com.hazelcast.client.config.ClientNetworkConfig;
import com.hazelcast.config.GroupConfig;
import com.hazelcast.core.HazelcastInstance;
import com.hazelcast.core.IMap;

/**
 * Unit tests of {@link DynamicTopologyListener}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class DynamicTopologyListenerTest extends AbstractRegistryOverlayImplTest {

    /**
     * Key of the topology inside the distributed map
     */
    private final static String TOPOLOGY_KEY = "my-topology";

    private final static String CONTAINER_NAME_1 = "sample-1";

    private final static String CONTAINER_DESCRIPTION_1 = "description of the container 1";

    private final static int CONTAINER_TCP_PORT_1 = 7801;

    private final LoggingUtil log = new LoggingUtil(
            java.util.logging.Logger.getLogger(DynamicTopologyListenerTest.class.getSimpleName()));

    private HazelcastInstance hzlClient;

    @Before
    public void before() {
        final ClientConfig hzlConfig = new ClientConfig();
        final ClientNetworkConfig networkConfig = new ClientNetworkConfig();
        networkConfig.addAddress(PETALS_REGISTRY_NODE.getHost() + ":" + PETALS_REGISTRY_NODE.getPort());
        hzlConfig.setNetworkConfig(networkConfig);
        final GroupConfig groupConfig = new GroupConfig();
        hzlConfig.setGroupConfig(groupConfig);
        groupConfig.setName(PETALS_REGISTRY_NODE.getGroupName());
        groupConfig.setPassword(PETALS_REGISTRY_NODE.getPassword());
        this.hzlClient = HazelcastClient.newHazelcastClient(hzlConfig);
    }

    @After
    public void after() {
        this.hzlClient.shutdown();
    }

    /**
     * <p>
     * Check {@link DynamicTopologyListener} against adding and removing container:
     * <ul>
     * <li>A basic sample topology is added to the distributed map,</li>
     * <li>A container is added to the initial topology.</li>
     * </ul>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the basic sample topology is correctly added to the distributed map,</li>
     * <li>when the topology is updated adding a new container, it is correctly added to the configuration service,</li>
     * </ul>
     * </p>
     */
    @Test
    public void testEntryUpdated_ContainerAdded() throws InterruptedException {
        final Topology initialTopology = ConfigurationTestUtils.createBasicSampleTopology();
        final Subdomain subdomain = initialTopology.getDomain().getSubDomain().get(0);
        final ContainerConfiguration currentContainerConfig = new ContainerConfiguration();
        ConfigurationServiceUtils.initializeContainerConfiguration(currentContainerConfig, subdomain.getContainer()
                .get(0), ConfigurationTestUtils.SUBDOMAIN_NAME_0);

        final DynamicTopologyListenerTestCfgSvc dtltcs = new DynamicTopologyListenerTestCfgSvc(true,
                currentContainerConfig, null);
        final DynamicTopologyListener dtl = new DynamicTopologyListener(dtltcs,
                ConfigurationTestUtils.SUBDOMAIN_NAME_0, this.log);

        final IMap<String, Topology> topologies = hzlClient.getMap("testEntryUpdated_ContainerAdded");
        topologies.addEntryListener(dtl, TOPOLOGY_KEY, true);

        topologies.put(TOPOLOGY_KEY, initialTopology);

        final Set<ContainerConfiguration> containers_1 = dtltcs.getContainersConfiguration();
        assertNotNull("Container list is null", containers_1);
        assertEquals("Unexpected remote container list size", 0, containers_1.size());

        // Add a container to the topology
        final Topology topology_2 = ConfigurationTestUtils.createBasicSampleTopology();
        final Container container2_1 = new Container();
        topology_2.getDomain().getSubDomain().get(0).getContainer().add(container2_1);
        container2_1.setName(CONTAINER_NAME_1);
        container2_1.setDescription(CONTAINER_DESCRIPTION_1);
        container2_1.setHost(ConfigurationTestUtils.CONTAINER_HOST);
        container2_1.setUser(ConfigurationTestUtils.CONTAINER_JMX_USER);
        container2_1.setPassword(ConfigurationTestUtils.CONTAINER_JMX_PWD);
        final JmxService jmxService1 = new JmxService();
        container2_1.setJmxService(jmxService1);
        final TransportService transportService1 = new TransportService();
        transportService1.setTcpPort(CONTAINER_TCP_PORT_1);
        container2_1.setTransportService(transportService1);

        topologies.put(TOPOLOGY_KEY, topology_2);
        synchronized (dtltcs) {
            dtltcs.wait(10000);
        }

        final Set<ContainerConfiguration> containers_2 = dtltcs.getContainersConfiguration();
        assertNotNull("Container list is null", containers_2);
        assertEquals("Unexpected container list size", 1, containers_2.size());
        final ContainerConfiguration container_2 = containers_2
                .toArray(new ContainerConfiguration[containers_2.size()])[0];
        assertEquals("Unexpected container", CONTAINER_NAME_1, container_2.getName());
    }

    /**
     * <p>
     * Check {@link DynamicTopologyListener} against adding and removing container:
     * <ul>
     * <li>A basic sample topology with 2 containers is added to the distributed map,</li>
     * <li>The remote container is removed from the initial topology.</li>
     * </ul>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the initial topology is correctly added to the distributed map,</li>
     * <li>when the topology is updated removing the remote container, it is correctly removed from the configuration
     * service,</li>
     * </ul>
     * </p>
     */
    @Test
    public void testEntryUpdated_ContainerRemoved() throws InterruptedException {
        final Topology initialTopology = ConfigurationTestUtils.createBasicSampleTopology();
        final Subdomain subdomain = initialTopology.getDomain().getSubDomain().get(0);
        final ContainerConfiguration currentContainerConfig = new ContainerConfiguration();
        ConfigurationServiceUtils.initializeContainerConfiguration(currentContainerConfig, subdomain.getContainer()
                .get(0), ConfigurationTestUtils.SUBDOMAIN_NAME_0);

        final Container remoteContainer = new Container();
        remoteContainer.setName(CONTAINER_NAME_1);
        remoteContainer.setDescription(CONTAINER_DESCRIPTION_1);
        remoteContainer.setHost(ConfigurationTestUtils.CONTAINER_HOST);
        remoteContainer.setUser(ConfigurationTestUtils.CONTAINER_JMX_USER);
        remoteContainer.setPassword(ConfigurationTestUtils.CONTAINER_JMX_PWD);
        final JmxService jmxService1 = new JmxService();
        remoteContainer.setJmxService(jmxService1);
        final TransportService transportService1 = new TransportService();
        transportService1.setTcpPort(CONTAINER_TCP_PORT_1);
        remoteContainer.setTransportService(transportService1);
        final ContainerConfiguration remoteContainerConfig = new ContainerConfiguration();
        ConfigurationServiceUtils.initializeContainerConfiguration(remoteContainerConfig, remoteContainer,
                ConfigurationTestUtils.SUBDOMAIN_NAME_0);

        final DynamicTopologyListenerTestCfgSvc dtltcs = new DynamicTopologyListenerTestCfgSvc(false,
                currentContainerConfig, remoteContainerConfig);
        final DynamicTopologyListener dtl = new DynamicTopologyListener(dtltcs,
                ConfigurationTestUtils.SUBDOMAIN_NAME_0, this.log);

        final IMap<String, Topology> topologies = hzlClient.getMap("testEntryUpdated_ContainerRemoved");
        topologies.addEntryListener(dtl, TOPOLOGY_KEY, true);

        subdomain.getContainer().add(remoteContainer);
        topologies.put(TOPOLOGY_KEY, initialTopology);

        final Set<ContainerConfiguration> containers_1 = dtltcs.getContainersConfiguration();
        assertNotNull("Container list is null", containers_1);
        assertEquals("Unexpected container list size", 1, containers_1.size());
        final ContainerConfiguration container_1 = containers_1
                .toArray(new ContainerConfiguration[containers_1.size()])[0];
        assertEquals("Unexpected container", CONTAINER_NAME_1, container_1.getName());

        // Remove the remote container from the topology
        final Topology topology_2 = ConfigurationTestUtils.createBasicSampleTopology();
        topologies.put(TOPOLOGY_KEY, topology_2);
        synchronized (dtltcs) {
            dtltcs.wait(10000);
        }

        final Set<ContainerConfiguration> containers_2 = dtltcs.getContainersConfiguration();
        assertNotNull("Container list is null", containers_2);
        assertEquals("Unexpected container list size", 0, containers_2.size());
    }
}
