/**
 * Copyright (c) 2013-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.registry.overlay;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.util.Arrays;
import java.util.List;
import java.util.Map;

import javax.xml.namespace.QName;

import org.junit.Test;
import org.ow2.petals.jmx.commons.PetalsEndpointRegistryServiceConstants;
import org.ow2.petals.microkernel.api.jbi.servicedesc.JBIServiceEndpointImpl;
import org.ow2.petals.microkernel.api.jbi.servicedesc.Location;
import org.ow2.petals.microkernel.api.jbi.servicedesc.PetalsServiceEndpoint;
import org.ow2.petals.microkernel.api.jbi.servicedesc.PetalsServiceEndpoint.EndpointType;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;
import com.ebmwebsourcing.easycommons.uuid.QualifiedUUIDGenerator;

/**
 * Unit tests of {@link RegistryOverlayImpl#getAllEndpoints()}, {@link RegistryOverlayImpl#getAllInternalEndpoints()}
 * and {@link RegistryOverlayImpl#getAllExternalEndpoints()}.
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class RegistryOverlayImpl_GetAllEndpoints_Test extends AbstractRegistryOverlayImplTest {

    /**
     * A namespace name
     */
    private final static String A_NAMESPACE = "http://petals.ow2.org/unit-tests";

    /**
     * A service name
     */
    private final static QName SERVICE_NAME_1 = new QName(A_NAMESPACE, "service-1");

    /**
     * A interface name
     */
    private final static QName INTERFACE_NAME_1 = new QName(A_NAMESPACE, "interface-1");

    /**
     * A local container name
     */
    private final static String LOCAL_CONTAINER_NAME = "local-container";

    /**
     * A remote container name
     */
    private final static String REMOTE_CONTAINER_NAME = "remote-container";

    /**
     * An internal endpoint name on a local container
     */
    private final static String LOCAL_INTERNAL_ENDPOINT_NAME;

    /**
     * An external endpoint name on a local container
     */
    private final static String LOCAL_EXTERNAL_ENDPOINT_NAME;

    /**
     * An internal endpoint name on a remote container
     */
    private final static String REMOTE_INTERNAL_ENDPOINT_NAME;

    /**
     * An external endpoint name on a remote container
     */
    private final static String REMOTE_EXTERNAL_ENDPOINT_NAME;

    static {
        final QualifiedUUIDGenerator uuidGenerator = new QualifiedUUIDGenerator("petals");
        LOCAL_INTERNAL_ENDPOINT_NAME = uuidGenerator.getNewID();
        LOCAL_EXTERNAL_ENDPOINT_NAME = uuidGenerator.getNewID();
        REMOTE_INTERNAL_ENDPOINT_NAME = uuidGenerator.getNewID();
        REMOTE_EXTERNAL_ENDPOINT_NAME = uuidGenerator.getNewID();
    }

    /**
     * <p>
     * Check the method {@link RegistryOverlayImpl#getAllEndpoints()},
     * {@link RegistryOverlayImpl#getAllInternalEndpoints()} and {@link RegistryOverlayImpl#getAllExternalEndpoints()} :
     * <ol>
     * <li>in a dedicated topology,</li>
     * <li>an internal endpoint is registered on a local container,</li>
     * <li>an external endpoint is registered on a local container,</li>
     * <li>an internal endpoint is registered on a remote container,</li>
     * <li>an external endpoint is registered on a remote container,</li>
     * </ol>
     * </p>
     * <p>
     * Expected result:
     * <ol>
     * <li>no error occurs</li>
     * <li>all endpoints are retrieved through {@link RegistryOverlayImpl#getAllEndpoints()}</li>
     * <li>all internal endpoints are retrieved through {@link RegistryOverlayImpl#getAllInternalEndpoints()}</li>
     * <li>all external endpoints are retrieved through {@link RegistryOverlayImpl#getAllExternalEndpoints()}</li>
     * </ol>
     * </p>
     */
    @Test
    public void testToGetEndpoints() throws Exception {

        final String topologyName = this.getClass().getSimpleName() + "_testToGetEndpoints";
        final RegistryOverlayImpl localRegistryOverlayImpl = new RegistryOverlayImpl();
        ReflectionHelper.setPrivateField(localRegistryOverlayImpl, "configurationService",
                RegistryOverlayImpl_GetAllEndpoints_Test.createCfgServiceMock(PETALS_REGISTRY_NODE.getGroupName(),
                        PETALS_REGISTRY_NODE.getPassword(), PETALS_REGISTRY_NODE.getHost(),
                        String.valueOf(PETALS_REGISTRY_NODE.getPort()), topologyName, LOCAL_CONTAINER_NAME));

        final RegistryOverlayImpl remoteRegistryOverlayImpl = new RegistryOverlayImpl();
        ReflectionHelper.setPrivateField(remoteRegistryOverlayImpl, "configurationService",
                RegistryOverlayImpl_GetAllEndpoints_Test.createCfgServiceMock(PETALS_REGISTRY_NODE.getGroupName(),
                        PETALS_REGISTRY_NODE.getPassword(), PETALS_REGISTRY_NODE.getHost(),
                        String.valueOf(PETALS_REGISTRY_NODE.getPort()), topologyName, REMOTE_CONTAINER_NAME));
        try {
            localRegistryOverlayImpl.start();
            remoteRegistryOverlayImpl.start();

            final Location localLocation = new Location();
            localLocation.setContainerName(LOCAL_CONTAINER_NAME);

            final JBIServiceEndpointImpl localInternalServiceEndpointMock = new JBIServiceEndpointImpl();
            localInternalServiceEndpointMock.setEndpointName(LOCAL_INTERNAL_ENDPOINT_NAME);
            localInternalServiceEndpointMock.setServiceName(SERVICE_NAME_1);
            localInternalServiceEndpointMock.setInterfacesName(Arrays.asList(new QName[] { INTERFACE_NAME_1 }));
            localInternalServiceEndpointMock.setType(EndpointType.INTERNAL);
            localInternalServiceEndpointMock.setLocation(localLocation);

            final JBIServiceEndpointImpl localExternalServiceEndpointMock = new JBIServiceEndpointImpl();
            localExternalServiceEndpointMock.setEndpointName(LOCAL_EXTERNAL_ENDPOINT_NAME);
            localExternalServiceEndpointMock.setServiceName(SERVICE_NAME_1);
            localExternalServiceEndpointMock.setInterfacesName(Arrays.asList(new QName[] { INTERFACE_NAME_1 }));
            localExternalServiceEndpointMock.setType(EndpointType.EXTERNAL);
            localExternalServiceEndpointMock.setLocation(localLocation);

            final JBIServiceEndpointImpl remoteInternalServiceEndpointMock = new JBIServiceEndpointImpl();
            remoteInternalServiceEndpointMock.setEndpointName(REMOTE_INTERNAL_ENDPOINT_NAME);
            remoteInternalServiceEndpointMock.setServiceName(SERVICE_NAME_1);
            remoteInternalServiceEndpointMock.setInterfacesName(Arrays.asList(new QName[] { INTERFACE_NAME_1 }));
            remoteInternalServiceEndpointMock.setType(EndpointType.INTERNAL);
            remoteInternalServiceEndpointMock.setLocation(localLocation);

            final JBIServiceEndpointImpl remoteExternalServiceEndpointMock = new JBIServiceEndpointImpl();
            remoteExternalServiceEndpointMock.setEndpointName(REMOTE_EXTERNAL_ENDPOINT_NAME);
            remoteExternalServiceEndpointMock.setServiceName(SERVICE_NAME_1);
            remoteExternalServiceEndpointMock.setInterfacesName(Arrays.asList(new QName[] { INTERFACE_NAME_1 }));
            remoteExternalServiceEndpointMock.setType(EndpointType.EXTERNAL);
            remoteExternalServiceEndpointMock.setLocation(localLocation);

            // Add the local internal endpoint
            localRegistryOverlayImpl.activateEndpoint(SERVICE_NAME_1, LOCAL_INTERNAL_ENDPOINT_NAME,
                    Arrays.asList(new QName[] { INTERFACE_NAME_1 }), null, localInternalServiceEndpointMock, false);

            // Add the local external endpoint
            localRegistryOverlayImpl.registerExternalEndpoint(localExternalServiceEndpointMock);

            // Add the remote internal endpoint
            remoteRegistryOverlayImpl.activateEndpoint(SERVICE_NAME_1, REMOTE_INTERNAL_ENDPOINT_NAME,
                    Arrays.asList(new QName[] { INTERFACE_NAME_1 }), null, remoteInternalServiceEndpointMock, false);

            // Add the remote external endpoint
            remoteRegistryOverlayImpl.registerExternalEndpoint(remoteExternalServiceEndpointMock);

            // Check that endpoints are registered
            final List<PetalsServiceEndpoint> registeredSvcEp = localRegistryOverlayImpl.getEndpoints();
            assertNotNull("Null list returned", registeredSvcEp);
            assertEquals("Invalid response number retrieving endpoints for service", registeredSvcEp.size(), 4);
            boolean localInternalEpRegistered = false;
            boolean localExternalEpRegistered = false;
            boolean remoteInternalEpRegistered = false;
            boolean remoteExternalEpRegistered = false;
            for (final PetalsServiceEndpoint svcEp : registeredSvcEp) {
                if (svcEp.getEndpointName().equals(LOCAL_INTERNAL_ENDPOINT_NAME)) {
                    localInternalEpRegistered = true;
                } else if (svcEp.getEndpointName().equals(LOCAL_EXTERNAL_ENDPOINT_NAME)) {
                    localExternalEpRegistered = true;
                } else if (svcEp.getEndpointName().equals(REMOTE_INTERNAL_ENDPOINT_NAME)) {
                    remoteInternalEpRegistered = true;
                } else if (svcEp.getEndpointName().equals(REMOTE_EXTERNAL_ENDPOINT_NAME)) {
                    remoteExternalEpRegistered = true;
                } else {
                    fail("Not activeated endpoint returned");
                }
            }
            assertTrue("The local internal endpoint is missing", localInternalEpRegistered);
            assertTrue("The local external endpoint is missing", localExternalEpRegistered);
            assertTrue("The remote internal endpoint is missing", remoteInternalEpRegistered);
            assertTrue("The remote external endpoint is missing", remoteExternalEpRegistered);

            // get all endpoints
            final List<Map<String, Object>> allEndpoints = localRegistryOverlayImpl.getAllEndpoints();
            assertNotNull("Null list returned", allEndpoints);
            assertEquals("No endpoint returned", allEndpoints.size(), 4);
            localInternalEpRegistered = false;
            localExternalEpRegistered = false;
            remoteInternalEpRegistered = false;
            remoteExternalEpRegistered = false;
            for (final Map<String, Object> endpoint : allEndpoints) {
                if (endpoint.get(PetalsEndpointRegistryServiceConstants.Endpoint.ENDPOINT_NAME).equals(
                        LOCAL_INTERNAL_ENDPOINT_NAME)) {
                    localInternalEpRegistered = true;
                } else if (endpoint.get(PetalsEndpointRegistryServiceConstants.Endpoint.ENDPOINT_NAME).equals(
                        LOCAL_EXTERNAL_ENDPOINT_NAME)) {
                    localExternalEpRegistered = true;
                } else if (endpoint.get(PetalsEndpointRegistryServiceConstants.Endpoint.ENDPOINT_NAME).equals(
                        REMOTE_INTERNAL_ENDPOINT_NAME)) {
                    remoteInternalEpRegistered = true;
                } else if (endpoint.get(PetalsEndpointRegistryServiceConstants.Endpoint.ENDPOINT_NAME).equals(
                        REMOTE_EXTERNAL_ENDPOINT_NAME)) {
                    remoteExternalEpRegistered = true;
                } else {
                    fail("Not activated endpoint returned");
                }
            }
            assertTrue("The local internal endpoint is missing", localInternalEpRegistered);
            assertTrue("The local external endpoint is missing", localExternalEpRegistered);
            assertTrue("The remote internal endpoint is missing", remoteInternalEpRegistered);
            assertTrue("The remote external endpoint is missing", remoteExternalEpRegistered);

            // get all internal endpoints
            final List<Map<String, Object>> allInternalEndpoints = localRegistryOverlayImpl.getAllInternalEndpoints();
            assertNotNull("Null list returned", allInternalEndpoints);
            assertEquals("Unexpected number of internal endpoints", allInternalEndpoints.size(), 2);
            localInternalEpRegistered = false;
            localExternalEpRegistered = false;
            remoteInternalEpRegistered = false;
            remoteExternalEpRegistered = false;
            for (final Map<String, Object> endpoint : allInternalEndpoints) {
                if (endpoint.get(PetalsEndpointRegistryServiceConstants.Endpoint.ENDPOINT_NAME).equals(
                        LOCAL_INTERNAL_ENDPOINT_NAME)) {
                    localInternalEpRegistered = true;
                } else if (endpoint.get(PetalsEndpointRegistryServiceConstants.Endpoint.ENDPOINT_NAME).equals(
                        LOCAL_EXTERNAL_ENDPOINT_NAME)) {
                    localExternalEpRegistered = true;
                } else if (endpoint.get(PetalsEndpointRegistryServiceConstants.Endpoint.ENDPOINT_NAME).equals(
                        REMOTE_INTERNAL_ENDPOINT_NAME)) {
                    remoteInternalEpRegistered = true;
                } else if (endpoint.get(PetalsEndpointRegistryServiceConstants.Endpoint.ENDPOINT_NAME).equals(
                        REMOTE_EXTERNAL_ENDPOINT_NAME)) {
                    remoteExternalEpRegistered = true;
                } else {
                    fail("Not activated endpoint returned");
                }
            }
            assertTrue("The local internal endpoint is missing", localInternalEpRegistered);
            assertFalse("The local external endpoint is returned", localExternalEpRegistered);
            assertTrue("The remote internal endpoint is missing", remoteInternalEpRegistered);
            assertFalse("The remote external endpoint is returned", remoteExternalEpRegistered);

            // get all external endpoints
            final List<Map<String, Object>> allExternalEndpoints = localRegistryOverlayImpl.getAllExternalEndpoints();
            assertNotNull("Null list returned", allExternalEndpoints);
            assertEquals("Unexpected numer of external endpoints", allExternalEndpoints.size(), 2);
            localInternalEpRegistered = false;
            localExternalEpRegistered = false;
            remoteInternalEpRegistered = false;
            remoteExternalEpRegistered = false;
            for (final Map<String, Object> endpoint : allExternalEndpoints) {
                if (endpoint.get(PetalsEndpointRegistryServiceConstants.Endpoint.ENDPOINT_NAME).equals(
                        LOCAL_INTERNAL_ENDPOINT_NAME)) {
                    localInternalEpRegistered = true;
                } else if (endpoint.get(PetalsEndpointRegistryServiceConstants.Endpoint.ENDPOINT_NAME).equals(
                        LOCAL_EXTERNAL_ENDPOINT_NAME)) {
                    localExternalEpRegistered = true;
                } else if (endpoint.get(PetalsEndpointRegistryServiceConstants.Endpoint.ENDPOINT_NAME).equals(
                        REMOTE_INTERNAL_ENDPOINT_NAME)) {
                    remoteInternalEpRegistered = true;
                } else if (endpoint.get(PetalsEndpointRegistryServiceConstants.Endpoint.ENDPOINT_NAME).equals(
                        REMOTE_EXTERNAL_ENDPOINT_NAME)) {
                    remoteExternalEpRegistered = true;
                } else {
                    fail("Not activated endpoint returned");
                }
            }
            assertFalse("The local internal endpoint is returned", localInternalEpRegistered);
            assertTrue("The local external endpoint is missing", localExternalEpRegistered);
            assertFalse("The remote internal endpoint is returned", remoteInternalEpRegistered);
            assertTrue("The remote external endpoint is missing", remoteExternalEpRegistered);

        } finally {
            remoteRegistryOverlayImpl.stop();
            localRegistryOverlayImpl.stop();
        }
    }

}
