/**
 * Copyright (c) 2014-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.registry.overlay;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.util.concurrent.Semaphore;
import java.util.concurrent.atomic.AtomicBoolean;

import org.junit.Test;
import org.ow2.petals.microkernel.api.jbi.messaging.registry.exception.DynamicTopologyNotLockedByCurrentThreadException;
import org.ow2.petals.microkernel.api.jbi.messaging.registry.exception.DynamicTopologyNotLockedException;
import org.ow2.petals.topology.generated.Container;
import org.ow2.petals.topology.generated.Domain;
import org.ow2.petals.topology.generated.Subdomain;
import org.ow2.petals.topology.generated.Topology;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;

/**
 * Unit tests of {@link RegistryOverlayImpl#setDynamicTopology(Topology)} and
 * {@link RegistryOverlayImpl#getDynamicTopology()}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class RegistryOverlayImpl_SetDynamicTopology_Test extends AbstractRegistryOverlayImplTest {

    /**
     * A container name
     */
    private final static String CONTAINER_NAME_1 = "container-1";

    /**
     * Another container name
     */
    private final static String CONTAINER_NAME_2 = "container-2";

    /**
     * <p>
     * Check the method {@link RegistryOverlayImpl#setDynamicTopology(Topology)}:
     * <ol>
     * <li>in a dedicated topology with two nodes,</li>
     * <li>no topology already exists,</li>
     * </ol>
     * </p>
     * <p>
     * Expected result:
     * <ol>
     * <li>the error {@link DynamicTopologyNotLockedException} occurs setting the dynamic topology when not locked
     * previously,</li>
     * <li>the error {@link DynamicTopologyNotLockedByCurrentThreadException} occurs setting the dynamic topology from a
     * thread that has not locked the topology objects,</li>
     * <li>no error occurs if the dynamic topology objects are locked previously,</li>
     * <li>the topology is set even if already set</li>
     * <li>the topology set on one node is correctly retrieved from the other Petals node</li>
     * </ol>
     * </p>
     */
    @Test
    public void testSetTopology() throws Exception {

        final Topology topology_1 = new Topology();
        final Domain domain_1 = new Domain();
        topology_1.setDomain(domain_1);
        final String expectedDomainName = "my-expected-domain-1";
        domain_1.setName(expectedDomainName);
        final String expectedDescription1 = "Description of domain '" + expectedDomainName
                + "' in topology 'topology_1'";
        domain_1.setDescription(expectedDescription1);
        final Subdomain subdomain_1 = new Subdomain();
        final String expectedSubdomainName = "my-sub-domain-1";
        subdomain_1.setName(expectedSubdomainName);
        domain_1.getSubDomain().add(subdomain_1);
        final Container container_1 = new Container();
        container_1.setName(CONTAINER_NAME_1);
        subdomain_1.getContainer().add(container_1);

        // The same topology as previous
        final Topology topology_2 = new Topology();
        final Domain domain_2 = new Domain();
        topology_2.setDomain(domain_2);
        domain_2.setName(expectedDomainName);
        final String expectedDescription2 = "Description of domain '" + expectedDomainName
                + "' in topology 'topology_2'";
        domain_2.setDescription(expectedDescription2);
        final Subdomain subdomain_2 = new Subdomain();
        subdomain_2.setName(expectedSubdomainName);
        domain_2.getSubDomain().add(subdomain_2);
        final Container container_2 = new Container();
        container_2.setName(CONTAINER_NAME_1);
        subdomain_2.getContainer().add(container_2);

        final RegistryOverlayImpl registryOverlayImpl_1 = new RegistryOverlayImpl();
        ReflectionHelper.setPrivateField(registryOverlayImpl_1, "configurationService",
                RegistryOverlayImpl_SetDynamicTopology_Test.createCfgServiceMock(AbstractRegistryOverlayImplTest
                        .createDomainConfiguration(expectedDomainName), AbstractRegistryOverlayImplTest
                        .createSubDomainConfiguration(PETALS_REGISTRY_NODE.getGroupName(),
                                PETALS_REGISTRY_NODE.getPassword(), PETALS_REGISTRY_NODE.getHost(),
                                String.valueOf(PETALS_REGISTRY_NODE.getPort()), expectedSubdomainName),
                        AbstractRegistryOverlayImplTest.createContainerConfiguration(CONTAINER_NAME_1)));

        final RegistryOverlayImpl registryOverlayImpl_2 = new RegistryOverlayImpl();
        ReflectionHelper.setPrivateField(registryOverlayImpl_2, "configurationService",
                RegistryOverlayImpl_SetDynamicTopology_Test.createCfgServiceMock(AbstractRegistryOverlayImplTest
                        .createDomainConfiguration(expectedDomainName), AbstractRegistryOverlayImplTest
                        .createSubDomainConfiguration(PETALS_REGISTRY_NODE.getGroupName(),
                                PETALS_REGISTRY_NODE.getPassword(), PETALS_REGISTRY_NODE.getHost(),
                                String.valueOf(PETALS_REGISTRY_NODE.getPort()), expectedSubdomainName),
                        AbstractRegistryOverlayImplTest.createContainerConfiguration(CONTAINER_NAME_2)));

        try {
            registryOverlayImpl_1.start();
            registryOverlayImpl_2.start();

            // /////////////////////////////////////////
            // Try to set a topology when not locked //
            // /////////////////////////////////////////
            try {
                registryOverlayImpl_1.setDynamicTopology(topology_1);
            } catch (final DynamicTopologyNotLockedException e) {
                // NOP, expected exception
            }

            // //////////////////////////////////////////////////////////
            // Try to set a topology when lock hold by another thread //
            // //////////////////////////////////////////////////////////
            final AtomicBoolean isCorrectlyLocked = new AtomicBoolean(false);
            final AtomicBoolean isErrorOccures = new AtomicBoolean(false);

            final Semaphore lockIsLockedLock = new Semaphore(1);
            final Semaphore waitBeforeUnLockByOwnerLock = new Semaphore(1);
            final Semaphore lockIsUnlockedLock = new Semaphore(1);

            lockIsLockedLock.acquire();
            lockIsUnlockedLock.acquire();
            final Thread aThread = new Thread(new Runnable() {

                @Override
                public void run() {

                    try {
                        final boolean isLockLocked = registryOverlayImpl_2.lockDynamicTopology();
                        if (isLockLocked) {
                            isCorrectlyLocked.set(true);
                        }
                        waitBeforeUnLockByOwnerLock.acquire();
                        lockIsLockedLock.release();

                        waitBeforeUnLockByOwnerLock.acquire();
                        registryOverlayImpl_2.unlockDynamicTopology();
                        lockIsUnlockedLock.release();

                    } catch (final InterruptedException e) {
                        isErrorOccures.set(true);
                        e.printStackTrace();
                    } catch (final DynamicTopologyNotLockedException e) {
                        isErrorOccures.set(true);
                        e.printStackTrace();
                    } catch (final DynamicTopologyNotLockedByCurrentThreadException e) {
                        isErrorOccures.set(true);
                        e.printStackTrace();
                    }
                }
            });

            aThread.start();
            try {
                lockIsLockedLock.acquire();
                assertTrue("Lock not correctly initialized, not locked", isCorrectlyLocked.get());

                try {
                    registryOverlayImpl_1.setDynamicTopology(topology_1);
                } catch (final DynamicTopologyNotLockedByCurrentThreadException e) {
                    // NOP, expected exception
                }
                waitBeforeUnLockByOwnerLock.release();
                lockIsUnlockedLock.acquire();
            } finally {
                aThread.interrupt();
            }

            // ////////////////////////////////////////////////
            // Try to set a topology with normal conditions //
            // ////////////////////////////////////////////////
            registryOverlayImpl_1.lockDynamicTopology();
            try {
                registryOverlayImpl_1.setDynamicTopology(topology_1);

                assertNotNull("null topology on node 1", registryOverlayImpl_1.getDynamicTopology());
                assertNotNull("null topology on node 2", registryOverlayImpl_2.getDynamicTopology());
                assertNotNull("null domain on node 1", registryOverlayImpl_1.getDynamicTopology().getDomain());
                assertNotNull("null domain on node 2", registryOverlayImpl_2.getDynamicTopology().getDomain());
                assertEquals(expectedDomainName, registryOverlayImpl_1.getDynamicTopology().getDomain().getName());
                assertEquals(expectedDescription1, registryOverlayImpl_1.getDynamicTopology().getDomain()
                        .getDescription());
                assertEquals(expectedDomainName, registryOverlayImpl_2.getDynamicTopology().getDomain().getName());
                assertEquals(expectedDescription1, registryOverlayImpl_2.getDynamicTopology().getDomain()
                        .getDescription());

                registryOverlayImpl_1.setDynamicTopology(topology_2);

                assertNotNull("null topology on node 1", registryOverlayImpl_1.getDynamicTopology());
                assertNotNull("null topology on node 2", registryOverlayImpl_2.getDynamicTopology());
                assertNotNull("null domain on node 1", registryOverlayImpl_1.getDynamicTopology().getDomain());
                assertNotNull("null domain on node 2", registryOverlayImpl_2.getDynamicTopology().getDomain());
                assertEquals(expectedDomainName, registryOverlayImpl_1.getDynamicTopology().getDomain().getName());
                assertEquals(expectedDescription2, registryOverlayImpl_1.getDynamicTopology().getDomain()
                        .getDescription());
                assertEquals(expectedDomainName, registryOverlayImpl_2.getDynamicTopology().getDomain().getName());
                assertEquals(expectedDescription2, registryOverlayImpl_2.getDynamicTopology().getDomain()
                        .getDescription());

            } finally {
                registryOverlayImpl_1.unlockDynamicTopology();
            }

        } finally {
            registryOverlayImpl_1.stop();
            registryOverlayImpl_2.stop();
        }
    }
}
